"""Tests for template tags in the projectroles app"""

import mistune
import uuid

from importlib import import_module

from django.conf import settings
from django.test import override_settings, RequestFactory
from django.urls import reverse, resolve

from test_plus.test import TestCase

import projectroles
from projectroles.app_settings import AppSettingAPI
from projectroles.models import AppSetting, SODAR_CONSTANTS
from projectroles.plugins import PluginAPI
from projectroles.templatetags import (
    projectroles_common_tags as c_tags,
    projectroles_role_tags as r_tags,
    projectroles_tags as tags,
)
from projectroles.tests.test_models import (
    ProjectMixin,
    RoleMixin,
    RoleAssignmentMixin,
    ProjectInviteMixin,
)


app_settings = AppSettingAPI()
plugin_api = PluginAPI()
site = import_module(settings.SITE_PACKAGE)


# SODAR constants
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_CONTRIBUTOR = SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR']
PROJECT_ROLE_GUEST = SODAR_CONSTANTS['PROJECT_ROLE_GUEST']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
SITE_MODE_SOURCE = SODAR_CONSTANTS['SITE_MODE_SOURCE']
SITE_MODE_TARGET = SODAR_CONSTANTS['SITE_MODE_TARGET']
SITE_MODE_PEER = SODAR_CONSTANTS['SITE_MODE_PEER']
REMOTE_LEVEL_READ_ROLES = SODAR_CONSTANTS['REMOTE_LEVEL_READ_ROLES']

# Local constants
APP_NAME_EX = 'example_project_app'
APP_NAME_FF = 'filesfolders'
NON_EXISTING_UUID = uuid.uuid4()
STATIC_FILE_PATH = 'images/logo_navbar.png'
TEMPLATE_PATH = 'projectroles/home.html'


class TemplateTagTestBase(
    ProjectMixin, RoleMixin, RoleAssignmentMixin, ProjectInviteMixin, TestCase
):
    """Base class for testing template tags"""

    def setUp(self):
        # Init roles
        self.init_roles()
        # Init users
        self.user = self.make_user('user_owner')
        self.superuser = self.make_user('user_superuser')
        self.superuser.is_superuser = True
        self.superuser.save()
        # Init category
        self.category = self.make_project(
            title='TestCategoryTop', type=PROJECT_TYPE_CATEGORY, parent=None
        )
        # Init project under category
        self.project = self.make_project(
            title='TestProjectSub',
            type=PROJECT_TYPE_PROJECT,
            parent=self.category,
        )
        # Init role assignments
        self.owner_as_cat = self.make_assignment(
            self.category, self.user, self.role_owner
        )
        self.owner_as = self.make_assignment(
            self.project, self.user, self.role_owner
        )
        # Init app_settings
        app_settings.set(
            APP_NAME_FF, 'allow_public_links', True, project=self.project
        )
        app_settings.set(
            'projectroles', 'ip_restrict', True, project=self.project
        )
        self.setting_filesfolders = AppSetting.objects.get(
            project=self.project,
            app_plugin__name=APP_NAME_FF,
            name='allow_public_links',
        )
        # Init request factory
        self.req_factory = RequestFactory()


class TestProjectrolesCommonTags(TemplateTagTestBase):
    """Tests for template tags in projectroles_common_tags"""

    def test_site_version(self):
        """Test site_version()"""
        self.assertEqual(c_tags.site_version(), site.__version__)

    def test_core_version(self):
        """Test core_version()"""
        self.assertEqual(c_tags.core_version(), projectroles.__version__)

    def test_check_backend(self):
        """Test check_backend()"""
        self.assertEqual(c_tags.check_backend('timeline_backend'), True)
        self.assertEqual(c_tags.check_backend('example_backend_app'), True)
        self.assertEqual(c_tags.check_backend('sodar_cache'), True)
        self.assertEqual(c_tags.check_backend('NON_EXISTING_PLUGIN'), False)

    def test_get_project_by_uuid(self):
        """Test get_project_by_uuid()"""
        self.assertEqual(
            c_tags.get_project_by_uuid(self.project.sodar_uuid), self.project
        )
        self.assertEqual(c_tags.get_project_by_uuid(NON_EXISTING_UUID), None)

    def test_get_user_by_uuid(self):
        """Test get_user_by_uuid()"""
        self.assertEqual(
            c_tags.get_user_by_uuid(self.user.sodar_uuid), self.user
        )
        self.assertEqual(c_tags.get_user_by_uuid(NON_EXISTING_UUID), None)

    def test_get_user_by_username(self):
        """Test get_user_by_username()"""
        self.assertEqual(
            c_tags.get_user_by_username(self.user.username), self.user
        )
        self.assertEqual(c_tags.get_user_by_username('NON_EXISTING_USER'), None)

    def test_get_django_setting(self):
        """Test get_django_setting()"""
        ret = c_tags.get_django_setting('PROJECTROLES_BROWSER_WARNING')
        self.assertEqual(ret, True)
        self.assertIsInstance(ret, bool)
        ret = c_tags.get_django_setting('PROJECTROLES_BROWSER_WARNING', js=True)
        self.assertEqual(ret, 1)
        self.assertIsInstance(ret, int)
        self.assertEqual(
            c_tags.get_django_setting('NON_EXISTING_SETTING'), None
        )
        self.assertEqual(
            c_tags.get_django_setting(
                'NON_EXISTING_SETTING', default='default'
            ),
            'default',
        )

    def test_get_app_setting(self):
        """Test get_app_setting()"""
        self.assertEqual(
            c_tags.get_app_setting(
                APP_NAME_EX, 'project_bool_setting', project=self.project
            ),
            False,
        )
        self.assertEqual(
            c_tags.get_app_setting(
                APP_NAME_EX, 'user_bool_setting', user=self.user
            ),
            False,
        )

    def test_static_file_exists(self):
        """Test static_file_exists()"""
        self.assertEqual(c_tags.static_file_exists(STATIC_FILE_PATH), True)
        self.assertEqual(
            c_tags.static_file_exists('NON_EXISTING_PATH/FILE.txt'), False
        )

    def test_template_exists(self):
        """Test template_exists()"""
        self.assertEqual(c_tags.template_exists(TEMPLATE_PATH), True)
        self.assertEqual(
            c_tags.template_exists('projectroles/NON_EXISTING_FILE.html'), False
        )

    def test_get_full_url(self):
        """Test get_full_url()"""
        url = reverse(
            'projectroles:detail', kwargs={'project': self.project.sodar_uuid}
        )

        with self.login(self.user):
            request = self.req_factory.get(url)
            self.assertEqual(
                c_tags.get_full_url(request, url),
                f'http://testserver/project/{self.project.sodar_uuid}',
            )

    def test_get_display_name(self):
        """Test get_display_name()"""
        self.assertEqual(
            c_tags.get_display_name('PROJECT', title=False, count=1), 'project'
        )
        self.assertEqual(
            c_tags.get_display_name('PROJECT', title=True, count=1), 'Project'
        )
        self.assertEqual(
            c_tags.get_display_name('PROJECT', title=False, count=2), 'projects'
        )
        self.assertEqual(
            c_tags.get_display_name('PROJECT', title=True, count=2), 'Projects'
        )
        self.assertEqual(
            c_tags.get_display_name('PROJECT', title=False, plural=True),
            'projects',
        )
        self.assertEqual(
            c_tags.get_display_name('PROJECT', title=True, plural=True),
            'Projects',
        )

    def test_get_project_title_html(self):
        """Test get_project_title_html()"""
        self.assertEqual(
            c_tags.get_project_title_html(self.project),
            'TestCategoryTop / TestProjectSub',
        )

    def test_get_project_link(self):
        """Test get_project_link()"""
        self.assertEqual(
            c_tags.get_project_link(self.project, full_title=False),
            f'<a href="/project/{self.project.sodar_uuid}" '
            f'title="{self.project.description}" data-toggle="tooltip" '
            f'data-placement="top">{self.project.title}</a>',
        )
        self.assertEqual(
            c_tags.get_project_link(self.project, full_title=True),
            f'<a href="/project/{self.project.sodar_uuid}" '
            f'title="{self.project.description}" data-toggle="tooltip" '
            f'data-placement="top">{self.project.full_title}</a>',
        )
        # TODO: Also test remote project link display (with icon)

    def test_get_user_superuser_icon(self):
        """Test get_user_superuser_icon()"""
        expected = (
            '<span title="Superuser" data-toggle="tooltip">'
            '<i class="iconify text-info ml-1" '
            'data-icon="mdi:shield-account"></i>'
            '</span>'
        )
        self.assertEqual(c_tags.get_user_superuser_icon(), expected)

    def test_get_user_superuser_icon_tooltip_false(self):
        """Test get_user_superuser_icon() with tooltip=False"""
        expected = (
            '<i class="iconify text-info ml-1" '
            'data-icon="mdi:shield-account"></i>'
        )
        self.assertEqual(c_tags.get_user_superuser_icon(False), expected)

    def test_get_user_inactive_icon(self):
        """Test get_user_inactive_icon()"""
        expected = (
            '<span title="Inactive" data-toggle="tooltip">'
            '<i class="iconify text-secondary ml-1" '
            'data-icon="mdi:account-off"></i>'
            '</span>'
        )
        self.assertEqual(c_tags.get_user_inactive_icon(), expected)

    def test_get_user_inactive_icon_tooltip_false(self):
        """Test get_user_inactive_icon() with tooltip=False"""
        expected = (
            '<i class="iconify text-secondary ml-1" '
            'data-icon="mdi:account-off"></i>'
        )
        self.assertEqual(c_tags.get_user_inactive_icon(False), expected)

    def test_get_user_html(self):
        """Test get_user_html()"""
        self.assertEqual(
            c_tags.get_user_html(self.user),
            f'<span class="sodar-user-html" data-toggle="tooltip" '
            f'data-placement="top" title="{self.user.get_full_name()}">'
            f'<a href="mailto:{self.user.email}">{self.user.username}</a>'
            f'</span>',
        )

    def test_get_user_html_superuser(self):
        """Test get_user_html() with superuser"""
        self.user.is_superuser = True
        self.assertEqual(
            c_tags.get_user_html(self.user),
            '<span class="sodar-user-html" data-toggle="tooltip" '
            'data-placement="top" title="{}"><a href="mailto:{}">'
            '{}</a><i class="iconify text-info ml-1" '
            'data-icon="mdi:shield-account"></i></span>'.format(
                self.user.get_full_name() + ' (superuser)',
                self.user.email,
                self.user.username,
            ),
        )

    def test_get_user_html_inactive(self):
        """Test get_user_html() with inactive user"""
        self.user.is_active = False
        self.assertEqual(
            c_tags.get_user_html(self.user),
            '<span class="sodar-user-html text-secondary" '
            'data-toggle="tooltip" data-placement="top" title="{}">{}'
            '<i class="iconify text-secondary ml-1" '
            'data-icon="mdi:account-off"></i></span>'.format(
                self.user.get_full_name() + ' (inactive)', self.user.username
            ),
        )

    def test_get_user_html_superuser_inactive(self):
        """Test get_user_html() with inactive superuser"""
        self.user.is_superuser = True
        self.user.is_active = False
        self.assertEqual(
            c_tags.get_user_html(self.user),
            '<span class="sodar-user-html text-secondary" '
            'data-toggle="tooltip" data-placement="top" title="{}">{}'
            '<i class="iconify text-info ml-1" '
            'data-icon="mdi:shield-account"></i>'
            '<i class="iconify text-secondary ml-1" '
            'data-icon="mdi:account-off"></i></span>'.format(
                self.user.get_full_name() + ' (inactive)', self.user.username
            ),
        )

    def test_get_user_html_no_email(self):
        """Test get_user_html() with no email on user"""
        self.user.email = ''
        self.assertEqual(
            c_tags.get_user_html(self.user),
            f'<span class="sodar-user-html" data-toggle="tooltip" '
            f'data-placement="top" title="{self.user.get_full_name()}">'
            f'{self.user.username}</span>',
        )

    def test_get_user_badge(self):
        """Test get_user_badge()"""
        expected = (
            f'<span class="badge badge-primary sodar-obj-badge '
            f'sodar-user-badge sodar-user-badge-active" '
            f'title="{self.user.get_full_name()}" '
            f'data-toggle="tooltip" '
            f'data-uuid="{self.user.sodar_uuid}">'
            f'<i class="iconify" data-icon="mdi:account"></i> '
            f'<a class="text-white" href="mailto:{self.user.email}">'
            f'{self.user.username}</a></span>'
        )
        self.assertEqual(c_tags.get_user_badge(self.user), expected)

    def test_get_user_badge_no_email(self):
        """Test get_user_badge() with no email"""
        self.user.email = None
        expected = (
            f'<span class="badge badge-primary sodar-obj-badge '
            f'sodar-user-badge sodar-user-badge-active" '
            f'title="{self.user.get_full_name()}" '
            f'data-toggle="tooltip" '
            f'data-uuid="{self.user.sodar_uuid}">'
            f'<i class="iconify" data-icon="mdi:account"></i> '
            f'{self.user.username}</span>'
        )
        self.assertEqual(c_tags.get_user_badge(self.user), expected)

    def test_get_user_badge_superuser(self):
        """Test get_user_badge() with superuser"""
        expected = (
            f'<span class="badge badge-primary sodar-obj-badge '
            f'sodar-user-badge sodar-user-badge-superuser" '
            f'title="{self.superuser.get_full_name()}" '
            f'data-toggle="tooltip" '
            f'data-uuid="{self.superuser.sodar_uuid}">'
            f'<i class="iconify" data-icon="mdi:shield-account"></i> '
            f'<a class="text-white" href="mailto:{self.superuser.email}">'
            f'{self.superuser.username}</a></span>'
        )
        self.assertEqual(c_tags.get_user_badge(self.superuser), expected)

    def test_get_user_badge_inactive(self):
        """Test get_user_badge() with inactive user"""
        self.user.is_active = False
        expected = (
            f'<span class="badge badge-secondary sodar-obj-badge '
            f'sodar-user-badge sodar-user-badge-inactive" '
            f'title="{self.user.get_full_name()}" '
            f'data-toggle="tooltip" '
            f'data-uuid="{self.user.sodar_uuid}">'
            f'<i class="iconify" data-icon="mdi:account-off"></i> '
            f'{self.user.username}</span>'
        )
        self.assertEqual(c_tags.get_user_badge(self.user), expected)

    def test_get_user_badge_extra_class(self):
        """Test get_user_badge() with extra_class"""
        expected = (
            f'<span class="badge badge-primary sodar-obj-badge '
            f'sodar-user-badge sodar-user-badge-active mr-1" '
            f'title="{self.user.get_full_name()}" '
            f'data-toggle="tooltip" '
            f'data-uuid="{self.user.sodar_uuid}">'
            f'<i class="iconify" data-icon="mdi:account"></i> '
            f'<a class="text-white" href="mailto:{self.user.email}">'
            f'{self.user.username}</a></span>'
        )
        self.assertEqual(
            c_tags.get_user_badge(self.user, extra_class='mr-1'), expected
        )

    def test_get_project_badge(self):
        """Test get_project_badge()"""
        url = reverse(
            'projectroles:detail', kwargs={'project': self.project.sodar_uuid}
        )
        expected = (
            f'<span class="badge badge-info sodar-obj-badge '
            f'sodar-project-badge" title="{self.project.full_title}" '
            f'data-toggle="tooltip">'
            f'<i class="iconify" data-icon="mdi:cube"></i> '
            f'<a href="{url}">{self.project.title}</a>'
            f'</span>'
        )
        self.assertEqual(c_tags.get_project_badge(self.project), expected)

    def test_get_project_badge_category(self):
        """Test get_project_badge() with category"""
        url = reverse(
            'projectroles:detail', kwargs={'project': self.category.sodar_uuid}
        )
        expected = (
            f'<span class="badge badge-info sodar-obj-badge '
            f'sodar-project-badge" title="{self.category.full_title}" '
            f'data-toggle="tooltip">'
            f'<i class="iconify" data-icon="mdi:rhombus-split"></i> '
            f'<a href="{url}">{self.category.title}</a>'
            f'</span>'
        )
        self.assertEqual(c_tags.get_project_badge(self.category), expected)

    def test_get_project_badge_can_view_false(self):
        """Test get_project_badge() with can_view=False"""
        expected = (
            f'<span class="badge badge-secondary sodar-obj-badge '
            f'sodar-project-badge" title="{self.project.full_title}" '
            f'data-toggle="tooltip">'
            f'<i class="iconify" data-icon="mdi:cube"></i> {self.project.title}'
            f'</span>'
        )
        self.assertEqual(
            c_tags.get_project_badge(self.project, can_view=False), expected
        )

    def test_get_project_badge_variant(self):
        """Test get_project_badge() with specified variant"""
        url = reverse(
            'projectroles:detail', kwargs={'project': self.project.sodar_uuid}
        )
        expected = (
            f'<span class="badge badge-danger sodar-obj-badge '
            f'sodar-project-badge" title="{self.project.full_title}" '
            f'data-toggle="tooltip">'
            f'<i class="iconify" data-icon="mdi:cube"></i> '
            f'<a href="{url}">{self.project.title}</a>'
            f'</span>'
        )
        self.assertEqual(
            c_tags.get_project_badge(self.project, variant='danger'), expected
        )

    def test_get_project_badge_variant_uppercase(self):
        """Test get_project_badge() with variant in uppercase"""
        url = reverse(
            'projectroles:detail', kwargs={'project': self.project.sodar_uuid}
        )
        expected = (
            f'<span class="badge badge-danger sodar-obj-badge '
            f'sodar-project-badge" title="{self.project.full_title}" '
            f'data-toggle="tooltip">'
            f'<i class="iconify" data-icon="mdi:cube"></i> '
            f'<a href="{url}">{self.project.title}</a>'
            f'</span>'
        )
        self.assertEqual(
            c_tags.get_project_badge(self.project, variant='DANGER'), expected
        )

    def test_get_project_badge_extra_class(self):
        """Test get_project_badge() with extra_class set"""
        url = reverse(
            'projectroles:detail', kwargs={'project': self.project.sodar_uuid}
        )
        expected = (
            f'<span class="badge badge-info sodar-obj-badge '
            f'sodar-project-badge mr-1" title="{self.project.full_title}" '
            f'data-toggle="tooltip">'
            f'<i class="iconify" data-icon="mdi:cube"></i> '
            f'<a href="{url}">{self.project.title}</a>'
            f'</span>'
        )
        self.assertEqual(
            c_tags.get_project_badge(self.project, extra_class='mr-1'), expected
        )

    def test_get_history_dropdown(self):
        """Test get_history_dropdown()"""
        url = reverse(
            'timeline:list_object',
            kwargs={
                'project': self.project.sodar_uuid,
                'object_model': self.user.__class__.__name__,
                'object_uuid': self.user.sodar_uuid,
            },
        )
        self.assertEqual(
            c_tags.get_history_dropdown(self.user, self.project),
            '<a class="dropdown-item sodar-pr-role-link-history" href="{}">\n'
            '<i class="iconify" data-icon="mdi:clock-time-eight-outline"></i> '
            'History</a>\n'.format(url),
        )

    def test_highlight_search_term(self):
        """Test highlight_search_term()"""
        item = 'Some Highlighted Text'
        term = 'highlight'
        self.assertEqual(
            c_tags.highlight_search_term(item, term),
            'Some <span class="sodar-search-highlight">Highlight</span>ed Text',
        )
        self.assertEqual(c_tags.highlight_search_term(item, ''), item)

    def test_get_info_link(self):
        """Test get_info_link()"""
        self.assertEqual(
            c_tags.get_info_link('content'),
            '<a class="sodar-info-link" tabindex="0" data-toggle="popover" '
            'data-trigger="focus" data-placement="top" data-content="content" >'
            '<i class="iconify text-info" data-icon="mdi:information"></i></a>',
        )
        self.assertEqual(
            c_tags.get_info_link('content', html=True),
            '<a class="sodar-info-link" tabindex="0" data-toggle="popover" '
            'data-trigger="focus" data-placement="top" data-content="content" '
            'data-html="true">'
            '<i class="iconify text-info" data-icon="mdi:information"></i></a>',
        )

    # TODO: Test get_remote_icon() (need to set up remote projects)

    def test_render_markdown(self):
        """Test render_markdown()"""
        raw_md = '**Some markdown**'
        self.assertEqual(
            c_tags.render_markdown(raw_md), mistune.markdown(raw_md)
        )

    def test_force_wrap(self):
        """Test force_wrap()"""
        s = 'sometext'
        s_space = 'some text'
        s_hyphen = 'some-text'
        self.assertEqual(c_tags.force_wrap(s, 4), 'some<wbr />text')
        self.assertEqual(c_tags.force_wrap(s_space, 4), s_space)
        self.assertEqual(c_tags.force_wrap(s_hyphen, 4), s_hyphen)

    def test_get_class(self):
        """Test get_class()"""
        self.assertEqual(c_tags.get_class(self.project), 'Project')
        self.assertEqual(c_tags.get_class(self.project, lower=True), 'project')

    def test_include_invalid_plugin(self):
        """Test get_backend_include() plugin checks"""
        self.assertEqual(
            c_tags.get_backend_include('NON_EXISTING_PLUGIN', 'js'), ''
        )
        # Testing a plugin which is not backend
        self.assertEqual(c_tags.get_backend_include(APP_NAME_FF, 'js'), '')

    def test_include_none_value(self):
        """Test get_backend_include() none attribute check"""
        # TODO: Replace with get_app_plugin once implemented for backend plugins
        backend_plugin = plugin_api.get_active_plugins('backend')[0]
        type(backend_plugin).javascript_url = None
        type(backend_plugin).css_url = None

        self.assertEqual(
            c_tags.get_backend_include(backend_plugin.name, 'js'), ''
        )
        self.assertEqual(
            c_tags.get_backend_include(backend_plugin.name, 'css'), ''
        )

    def test_include_invalid_url(self):
        """Test get_backend_include() file existence check"""
        # TODO: Replace with get_app_plugin once implemented for backend plugins
        backend_plugin = plugin_api.get_active_plugins('backend')[0]

        type(backend_plugin).javascript_url = (
            'example_backend_app/js/NOT_EXISTING_JS.js'
        )
        type(backend_plugin).css_url = (
            'example_backend_app/css/NOT_EXISTING_CSS.css'
        )

        self.assertEqual(
            c_tags.get_backend_include(backend_plugin.name, 'js'), ''
        )
        self.assertEqual(
            c_tags.get_backend_include(backend_plugin.name, 'css'), ''
        )

    def test_get_backend_include(self):
        """Test get_backend_include()"""
        # TODO: Replace with get_app_plugin once implemented for backend plugins
        backend_plugin = plugin_api.get_active_plugins('backend')[0]

        type(backend_plugin).javascript_url = (
            'example_backend_app/js/greeting.js'
        )
        type(backend_plugin).css_url = 'example_backend_app/css/greeting.css'

        self.assertEqual(
            c_tags.get_backend_include(backend_plugin.name, 'js'),
            '<script type="text/javascript" '
            'src="/static/example_backend_app/js/greeting.js"></script>',
        )
        self.assertEqual(
            c_tags.get_backend_include(backend_plugin.name, 'css'),
            '<link rel="stylesheet" type="text/css" '
            'href="/static/example_backend_app/css/greeting.css"/>',
        )

    def test_split(self):
        """Test split()"""
        s = 'xyz'
        self.assertEqual(c_tags.split(s, 'y'), ['x', 'z'])


class TestProjectrolesRoleTags(TemplateTagTestBase):
    """Tests for template tags in projectroles_role_tags"""

    def test_get_role_icon(self):
        """Test get_role_icon()"""
        self.assertEqual(r_tags.get_role_icon(self.role_owner), 'mdi:star')
        self.assertEqual(
            r_tags.get_role_icon(self.role_delegate), 'mdi:star-half-full'
        )
        self.assertEqual(
            r_tags.get_role_icon(self.role_contributor), 'mdi:account'
        )
        self.assertEqual(r_tags.get_role_icon(self.role_guest), 'mdi:account')

    def test_get_role_class(self):
        """Test get_role_class()"""
        self.assertEqual(r_tags.get_role_class(self.user), '')

    def test_get_role_class_inactive(self):
        """Test get_role_class() with inactive user"""
        self.user.is_active = False
        self.assertEqual(r_tags.get_role_class(self.user), 'text-secondary')

    def test_get_role_perms(self):
        """Test get_role_perms()"""
        self.assertEqual(
            r_tags.get_role_perms(self.project, self.user),
            {
                'can_update_owner': True,
                'can_update_delegate': True,
                'can_update_members': True,
                'can_invite': True,
            },
        )
        # Guest
        user_guest = self.make_user('user_new')
        self.make_assignment(self.project, user_guest, self.role_guest)
        self.assertEqual(
            r_tags.get_role_perms(self.project, user_guest),
            {
                'can_update_owner': False,
                'can_update_delegate': False,
                'can_update_members': False,
                'can_invite': False,
            },
        )
        # Inherited delegate
        user_delegate_cat = self.make_user('user_delegate_cat')
        self.make_assignment(
            self.category, user_delegate_cat, self.role_delegate
        )
        self.assertEqual(
            r_tags.get_role_perms(self.project, user_delegate_cat),
            {
                'can_update_owner': False,
                'can_update_delegate': False,
                'can_update_members': True,
                'can_invite': True,
            },
        )

    def test_display_role_buttons(self):
        """Test display_role_buttons()"""
        self.assertTrue(
            r_tags.display_role_buttons(
                self.project,
                self.owner_as,
                r_tags.get_role_perms(self.project, self.user),
            ),
        )
        # Guest
        user_guest = self.make_user('user_new')
        guest_as = self.make_assignment(
            self.project, user_guest, self.role_guest
        )
        self.assertFalse(
            r_tags.display_role_buttons(
                self.project,
                guest_as,
                r_tags.get_role_perms(self.project, guest_as.user),
            ),
        )

    def test_get_inactive_role(self):
        """Test get_inactive_role()"""
        inh_user = self.make_user('inh_user')
        self.owner_as_cat.user = inh_user
        self.owner_as_cat.save()
        self.assertEqual(
            r_tags.get_inactive_role(self.project, self.owner_as_cat), None
        )
        inactive_as = self.make_assignment(
            self.project, inh_user, self.role_contributor
        )
        self.assertEqual(
            r_tags.get_inactive_role(self.project, self.owner_as_cat),
            inactive_as,
        )


class TestProjectrolesTags(TemplateTagTestBase):
    """Tests for template tags in projectroles_tags"""

    def test_sodar_constant(self):
        """Test sodar_constant()"""
        self.assertEqual(tags.sodar_constant('PROJECT_TYPE_PROJECT'), 'PROJECT')
        self.assertEqual(tags.sodar_constant('NON_EXISTING_CONSTANT'), None)

    def test_get_backend_plugins(self):
        """Test get_backend_plugins()"""
        self.assertEqual(
            len(tags.get_backend_plugins()),
            len(settings.ENABLED_BACKEND_PLUGINS),
        )

    # TODO: Test get_site_app_messages() (set up admin alert)

    def test_allow_project_creation(self):
        """Test allow_project_creation()"""
        self.assertEqual(tags.allow_project_creation(), True)

    @override_settings(
        PROJECTROLES_SITE_MODE='TARGET', PROJECTROLES_TARGET_CREATE=False
    )
    def test_allow_project_creation_target(self):
        """Test allow_project_creation() in target mode"""
        self.assertEqual(tags.allow_project_creation(), False)

    def test_is_app_visible(self):
        """Test is_app_visible()"""
        app_plugin = plugin_api.get_app_plugin(APP_NAME_FF)
        self.assertEqual(
            tags.is_app_visible(app_plugin, self.project, self.user), True
        )

    def test_is_app_visible_category(self):
        """Test is_app_visible() with a category"""
        app_plugin = plugin_api.get_app_plugin(APP_NAME_FF)
        self.assertEqual(
            tags.is_app_visible(app_plugin, self.category, self.user),
            False,
        )

    def test_is_app_visible_category_enabled(self):
        """Test is_app_visible() with category_enable=True"""
        app_plugin = plugin_api.get_app_plugin('timeline')
        self.assertEqual(
            tags.is_app_visible(app_plugin, self.category, self.user), True
        )

    @override_settings(PROJECTROLES_HIDE_PROJECT_APPS=[APP_NAME_FF])
    def test_is_app_visible_hide(self):
        """Test is_app_visible() with a hidden app and normal/superuser"""
        app_plugin = plugin_api.get_app_plugin(APP_NAME_FF)
        superuser = self.make_user('superuser')
        superuser.is_superuser = True
        superuser.save()
        self.assertEqual(
            tags.is_app_visible(app_plugin, self.project, self.user), False
        )
        self.assertEqual(
            tags.is_app_visible(app_plugin, self.project, superuser), False
        )

    def test_get_app_link_state(self):
        """Test get_app_link_state()"""
        app_plugin = plugin_api.get_app_plugin(APP_NAME_FF)
        # TODO: Why does this also require app_name?
        self.assertEqual(
            tags.get_app_link_state(app_plugin, APP_NAME_FF, 'list'),
            'active',
        )
        self.assertEqual(
            tags.get_app_link_state(
                app_plugin, APP_NAME_FF, 'NON_EXISTING_URL_NAME'
            ),
            '',
        )

    # TODO: Test get_pr_link_state()

    def test_get_help_highlight(self):
        """Test get_help_highlight()"""
        self.assertEqual(
            tags.get_help_highlight(self.user), 'font-weight-bold text-warning'
        )

    # TODO: Test get_role_import_action() (Set up remote projects)
    # TODO: Test get_target_project_select() (Set up remote projects)

    def test_get_remote_access_legend(self):
        """Test get_remote_access_legend()"""
        self.assertEqual(tags.get_remote_access_legend('NONE'), 'No access')
        self.assertEqual(
            tags.get_remote_access_legend('NON_EXISTING_LEVEL'), 'N/A'
        )

    def test_get_sidebar_app_legend(self):
        """Test get_sidebar_app_legend()"""
        self.assertEqual(
            tags.get_sidebar_app_legend('Update Project'), 'Update<br />Project'
        )

    def test_get_sidebar_links_home(self):
        """Test get_sidebar_links() on the home view"""
        url = reverse('home')
        request = self.req_factory.get(url)
        request.resolver_match = resolve(url)
        request.user = self.user
        self.assertEqual(
            tags.get_project_app_links(request),
            [],
        )

    def test_get_sidebar_links_project_detail_view(self):
        """Test get_sidebar_links() on project detail view"""
        url = reverse(
            'projectroles:detail', kwargs={'project': self.project.sodar_uuid}
        )
        request = self.req_factory.get(url)
        request.resolver_match = resolve(url)
        request.user = self.user
        self.assertEqual(
            tags.get_project_app_links(request, self.project),
            [
                {
                    'name': 'project-detail',
                    'url': f'/project/{self.project.sodar_uuid}',
                    'label': 'Project Overview',
                    'icon': 'mdi:cube',
                    'active': True,
                },
                {
                    'name': 'app-plugin-filesfolders',
                    'url': f'/files/{self.project.sodar_uuid}',
                    'label': 'Files',
                    'icon': 'mdi:file',
                    'active': False,
                },
                {
                    'name': 'app-plugin-timeline',
                    'url': f'/timeline/{self.project.sodar_uuid}',
                    'label': 'Timeline',
                    'icon': 'mdi:clock-time-eight',
                    'active': False,
                },
                {
                    'name': 'app-plugin-bgjobs',
                    'url': f'/bgjobs/list/{self.project.sodar_uuid}',
                    'label': 'Background Jobs',
                    'icon': 'mdi:server',
                    'active': False,
                },
                {
                    'name': 'app-plugin-example_project_app',
                    'url': f'/examples/project/{self.project.sodar_uuid}',
                    'label': 'Example Project App',
                    'icon': 'mdi:rocket-launch',
                    'active': False,
                },
                {
                    'name': 'project-roles',
                    'url': f'/project/members/{self.project.sodar_uuid}',
                    'label': 'Members',
                    'icon': 'mdi:account-multiple',
                    'active': False,
                },
                {
                    'name': 'project-update',
                    'url': f'/project/update/{self.project.sodar_uuid}',
                    'label': 'Update Project',
                    'icon': 'mdi:lead-pencil',
                    'active': False,
                },
            ],
        )

    def test_get_sidebar_links_role_view(self):
        """Test get_sidebar_links() on the role view"""
        url = reverse(
            'projectroles:roles', kwargs={'project': self.project.sodar_uuid}
        )
        request = self.req_factory.get(url)
        request.resolver_match = resolve(url)
        request.user = self.user
        for app in tags.get_project_app_links(request, self.project):
            if app['name'] == 'project-roles':
                self.assertEqual(app['active'], True)
            else:
                self.assertEqual(app['active'], False)

    def test_get_sidebar_links_timeline_view(self):
        """Test get_sidebar_links() on the timeline view"""
        url = reverse(
            'timeline:list_project', kwargs={'project': self.project.sodar_uuid}
        )
        request = self.req_factory.get(url)
        request.resolver_match = resolve(url)
        request.user = self.user
        for app in tags.get_project_app_links(request, self.project):
            if app['name'] == 'app-plugin-timeline':
                self.assertEqual(app['active'], True)
            else:
                self.assertEqual(app['active'], False)

    def test_get_user_links_home(self):
        """Test get_user_links() on the home view"""
        url = reverse('home')
        request = self.req_factory.get(url)
        request.resolver_match = resolve(url)
        request.user = self.user
        self.assertEqual(
            tags.get_user_links(request),
            [
                {
                    'name': 'appalerts',
                    'url': '/alerts/app/list',
                    'label': 'App Alerts',
                    'icon': 'mdi:alert-octagram',
                    'active': False,
                },
                {
                    'name': 'example_site_app',
                    'url': '/examples/site/example',
                    'label': 'Example Site App',
                    'icon': 'mdi:rocket-launch-outline',
                    'active': False,
                },
                {
                    'name': 'timeline_site',
                    'url': '/timeline/site',
                    'label': 'Site-Wide Events',
                    'icon': 'mdi:clock-time-eight-outline',
                    'active': False,
                },
                {
                    'name': 'tokens',
                    'url': '/tokens/',
                    'label': 'API Tokens',
                    'icon': 'mdi:key-chain-variant',
                    'active': False,
                },
                {
                    'name': 'userprofile',
                    'url': '/user/profile',
                    'label': 'User Profile',
                    'icon': 'mdi:account-details',
                    'active': False,
                },
                {
                    'name': 'sign-out',
                    'url': '/logout/',
                    'label': 'Logout',
                    'icon': 'mdi:logout-variant',
                    'active': False,
                },
            ],
        )

    def test_get_user_links_home_superuser(self):
        """Test get_user_links() on the home view as superuser"""
        url = reverse('home')
        request = self.req_factory.get(url)
        request.resolver_match = resolve(url)
        request.user = self.superuser
        self.assertEqual(
            tags.get_user_links(request),
            [
                {
                    'name': 'adminalerts',
                    'url': '/alerts/adm/list',
                    'label': 'Admin Alerts',
                    'icon': 'mdi:alert',
                    'active': False,
                },
                {
                    'name': 'appalerts',
                    'url': '/alerts/app/list',
                    'label': 'App Alerts',
                    'icon': 'mdi:alert-octagram',
                    'active': False,
                },
                {
                    'name': 'bgjobs_site',
                    'url': '/bgjobs/list',
                    'label': 'Site Background Jobs',
                    'icon': 'mdi:server',
                    'active': False,
                },
                {
                    'name': 'example_site_app',
                    'url': '/examples/site/example',
                    'label': 'Example Site App',
                    'icon': 'mdi:rocket-launch-outline',
                    'active': False,
                },
                {
                    'name': 'remotesites',
                    'url': '/project/remote/sites',
                    'label': 'Remote Site Access',
                    'icon': 'mdi:cloud-sync',
                    'active': False,
                },
                {
                    'name': 'siteappsettings',
                    'url': '/project/site-app-settings',
                    'label': 'Site App Settings',
                    'icon': 'mdi:cog-outline',
                    'active': False,
                },
                {
                    'name': 'siteinfo',
                    'url': '/siteinfo/info',
                    'label': 'Site Info',
                    'icon': 'mdi:bar-chart',
                    'active': False,
                },
                {
                    'name': 'timeline_site',
                    'url': '/timeline/site',
                    'label': 'Site-Wide Events',
                    'icon': 'mdi:clock-time-eight-outline',
                    'active': False,
                },
                {
                    'name': 'timeline_site_admin',
                    'url': '/timeline/site/all',
                    'label': 'All Timeline Events',
                    'icon': 'mdi:web-clock',
                    'active': False,
                },
                {
                    'name': 'tokens',
                    'url': '/tokens/',
                    'label': 'API Tokens',
                    'icon': 'mdi:key-chain-variant',
                    'active': False,
                },
                {
                    'name': 'userprofile',
                    'url': '/user/profile',
                    'label': 'User Profile',
                    'icon': 'mdi:account-details',
                    'active': False,
                },
                {
                    'name': 'admin',
                    'url': None,  # No URL for Django admin, opens warning modal
                    'label': 'Django Admin',
                    'icon': 'mdi:cogs',
                    'active': False,
                },
                {
                    'name': 'sign-out',
                    'url': '/logout/',
                    'label': 'Logout',
                    'icon': 'mdi:logout-variant',
                    'active': False,
                },
            ],
        )

    def test_get_user_links_userprofile(self):
        """Test get_user_links() on the user profile view"""
        url = reverse('userprofile:detail')
        request = self.req_factory.get(url)
        request.resolver_match = resolve(url)
        request.user = self.user
        for app in tags.get_user_links(request):
            if app['name'] == 'userprofile':
                self.assertEqual(app['active'], True)
            else:
                self.assertEqual(app['active'], False)

    def test_get_user_links_remote_site(self):
        """Test get_user_links() on the remote site view"""
        url = reverse('projectroles:remote_site_create')
        request = self.req_factory.get(url)
        request.resolver_match = resolve(url)
        request.user = self.user
        for app in tags.get_user_links(request):
            if app['name'] == 'remote_site_app':
                self.assertEqual(app['active'], True)
            else:
                self.assertEqual(app['active'], False)



# Generated by Django 3.1.8 on 2021-04-16 12:39

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('djangoplugins', '0001_initial'),
        ('projectroles', '0019_project_public_guest_access'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='AppAlert',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'alert_name',
                    models.CharField(
                        help_text='Internal alert name string', max_length=255
                    ),
                ),
                (
                    'message',
                    models.TextField(
                        help_text='Alert message (may contain HTML)'
                    ),
                ),
                (
                    'level',
                    models.CharField(
                        choices=[
                            ('INFO', 'Info'),
                            ('SUCCESS', 'Success'),
                            ('WARNING', 'Warning'),
                            ('DANGER', 'Danger'),
                        ],
                        default='INFO',
                        help_text='Alert level',
                        max_length=64,
                    ),
                ),
                (
                    'active',
                    models.BooleanField(
                        default=True, help_text='Active status of the alert'
                    ),
                ),
                (
                    'url',
                    models.URLField(
                        blank=True,
                        help_text='URL for the source of the alert (optional)',
                        max_length=2000,
                        null=True,
                    ),
                ),
                (
                    'date_created',
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text='DateTime of the alert creation',
                    ),
                ),
                (
                    'sodar_uuid',
                    models.UUIDField(
                        default=uuid.uuid4,
                        help_text='Alert SODAR UUID',
                        unique=True,
                    ),
                ),
                (
                    'app_plugin',
                    models.ForeignKey(
                        help_text='App to which the alert belongs',
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='app_alerts',
                        to='djangoplugins.plugin',
                    ),
                ),
                (
                    'project',
                    models.ForeignKey(
                        help_text='Project to which the alert belongs '
                        '(optional)',
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='app_alerts',
                        to='projectroles.project',
                    ),
                ),
                (
                    'user',
                    models.ForeignKey(
                        help_text='User who receives the alert',
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='app_alerts',
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
    ]



"""Deletecache management command for deleting SODARCache data"""

import sys

from django.conf import settings
from django.core.management.base import BaseCommand  # , CommandError

# Projectroles dependency
from projectroles.management.logging import ManagementCommandLogger
from projectroles.models import Project

from sodarcache.models import JSONCacheItem

logger = ManagementCommandLogger(__name__)


class Command(BaseCommand):
    help = 'Deletes cached data from external services'

    def add_arguments(self, parser):
        parser.add_argument(
            '-p',
            '--project',
            metavar='UUID',
            type=str,
            help='Limit deletion to a project',
        )

    def handle(self, *args, **options):
        if 'sodar_cache' not in settings.ENABLED_BACKEND_PLUGINS:
            logger.error(
                'SodarCache backend not enabled in settings, cancelled!'
            )
            sys.exit(1)

        project = None
        if options.get('project'):
            try:
                project = Project.objects.get(sodar_uuid=options['project'])
                logger.info(
                    f'Limiting deletion to project {project.get_log_title()}"'
                )
                items = JSONCacheItem.objects.filter(project=project)
            except Project.DoesNotExist:
                logger.error(
                    'Project not found with UUID={}'.format(options['project'])
                )
                sys.exit(1)
        else:
            items = JSONCacheItem.objects.all()

        item_count = items.count()
        if item_count == 0:
            logger.info('No cached data found')
            sys.exit(0)

        items.delete()
        logger.info(
            'Deleted {} cached data item{} from {}'.format(
                item_count,
                's' if item_count != 1 else '',
                (
                    'project {}'.format(project.get_log_title())
                    if project
                    else 'all projects'
                ),
            )
        )



"""UI view tests for the tokens app"""

from django.contrib.messages import get_messages
from django.test import override_settings
from django.urls import reverse
from django.utils import timezone

from knox.models import AuthToken

from test_plus.test import TestCase

# Projectroles dependency
from projectroles.models import SODAR_CONSTANTS
from projectroles.tests.test_models import (
    ProjectMixin,
    RoleMixin,
    RoleAssignmentMixin,
)

from tokens.views import (
    TOKEN_CREATE_MSG,
    TOKEN_DELETE_MSG,
    TOKEN_CREATE_RESTRICT_MSG,
)


# SODAR constants
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']


class TestUserTokenListView(
    ProjectMixin, RoleMixin, RoleAssignmentMixin, TestCase
):
    """Tests for UserTokenListView"""

    def _make_token(self):
        self.tokens = [AuthToken.objects.create(self.regular_user, None)]

    def setUp(self):
        self.regular_user = self.make_user('regular_user')
        self.url = reverse('tokens:list')

    def test_get(self):
        """Test UserTokenListView GET with token"""
        self._make_token()
        with self.login(self.regular_user):
            response = self.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.context['object_list']), 1)
        self.assertEqual(response.context['token_create_enable'], True)
        self.assertEqual(response.context['token_create_msg'], '')

    def test_get_no_tokens(self):
        """Test GET with no tokens"""
        with self.login(self.regular_user):
            response = self.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.context['object_list']), 0)

    @override_settings(TOKENS_CREATE_PROJECT_USER_RESTRICT=True)
    def test_get_restrict(self):
        """Test GET with restriction"""
        with self.login(self.regular_user):
            response = self.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context['token_create_enable'], False)
        self.assertEqual(
            response.context['token_create_msg'], TOKEN_CREATE_RESTRICT_MSG
        )

    @override_settings(TOKENS_CREATE_PROJECT_USER_RESTRICT=True)
    def test_get_restrict_role(self):
        """Test GET with restriction and user with role"""
        self.init_roles()
        category = self.make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self.make_assignment(category, self.regular_user, self.role_guest)
        with self.login(self.regular_user):
            response = self.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context['token_create_enable'], True)
        self.assertEqual(response.context['token_create_msg'], '')

    @override_settings(TOKENS_CREATE_PROJECT_USER_RESTRICT=True)
    def test_get_restrict_superuser(self):
        """Test GET with restriction as superuser"""
        superuser = self.make_user('superuser')
        superuser.is_superuser = True
        superuser.save()
        with self.login(superuser):
            response = self.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context['token_create_enable'], True)
        self.assertEqual(response.context['token_create_msg'], '')


class TestUserTokenCreateView(TestCase):
    """Tests for UserTokenCreateView"""

    def setUp(self):
        self.user = self.make_user()
        self.url = reverse('tokens:create')

    def test_get(self):
        """Test UserTokenCreateView GET"""
        with self.login(self.user):
            response = self.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertIsNotNone(response.context['form'])

    def test_post_no_ttl(self):
        """Test POST with TTL=0"""
        self.assertEqual(AuthToken.objects.count(), 0)
        with self.login(self.user):
            response = self.post(self.url, data={'ttl': 0})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            list(get_messages(response.wsgi_request))[0].message,
            TOKEN_CREATE_MSG,
        )
        self.assertEqual(AuthToken.objects.count(), 1)
        self.assertEqual(AuthToken.objects.first().expiry, None)

    def test_post_ttl(self):
        """Test POST with TTL != 0"""
        self.assertEqual(AuthToken.objects.count(), 0)
        with self.login(self.user):
            response = self.post(self.url, data={'ttl': 10})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            list(get_messages(response.wsgi_request))[0].message,
            TOKEN_CREATE_MSG,
        )
        self.assertEqual(AuthToken.objects.count(), 1)
        self.assertEqual(
            AuthToken.objects.first().expiry.replace(
                minute=0, second=0, microsecond=0
            ),
            (timezone.now() + timezone.timedelta(hours=10)).replace(
                minute=0, second=0, microsecond=0
            ),
        )


class TestUserTokenDeleteView(TestCase):
    """Tests for UserTokenDeleteView"""

    def setUp(self):
        self.user = self.make_user()
        self.token = AuthToken.objects.create(user=self.user)[0]
        self.url = reverse('tokens:delete', kwargs={'pk': self.token.pk})

    def test_get(self):
        """Test UserTokenDeleteView GET"""
        with self.login(self.user):
            response = self.get(self.url)
        self.assertEqual(response.status_code, 200)

    def test_post(self):
        """Test token deletion"""
        self.assertEqual(AuthToken.objects.count(), 1)
        with self.login(self.user):
            response = self.post(self.url)
            self.assertRedirects(
                response, reverse('tokens:list'), fetch_redirect_response=False
            )
        self.assertEqual(
            list(get_messages(response.wsgi_request))[0].message,
            TOKEN_DELETE_MSG,
        )
        self.assertEqual(AuthToken.objects.count(), 0)



import rules

from django.conf import settings

from projectroles.app_settings import AppSettingAPI
from projectroles.models import RoleAssignment, SODAR_CONSTANTS


app_settings = AppSettingAPI()


# SODAR constants
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_CONTRIBUTOR = SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR']
PROJECT_ROLE_GUEST = SODAR_CONSTANTS['PROJECT_ROLE_GUEST']
PROJECT_ROLE_VIEWER = SODAR_CONSTANTS['PROJECT_ROLE_VIEWER']
PROJECT_ROLE_FINDER = SODAR_CONSTANTS['PROJECT_ROLE_FINDER']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
SITE_MODE_SOURCE = SODAR_CONSTANTS['SITE_MODE_SOURCE']
SITE_MODE_TARGET = SODAR_CONSTANTS['SITE_MODE_TARGET']

# Local constants
APP_NAME = 'projectroles'


# Predicates -------------------------------------------------------------------


@rules.predicate
def is_project_owner(user, obj):
    """
    Whether or not the user has the role of project owner, or is the owner of
    a parent category of the current project.
    """
    return obj.is_owner(user) if obj else False


@rules.predicate
def is_project_delegate(user, obj):
    """Whether or not the user has the role of project delegate"""
    return obj.is_delegate(user) if obj else False


@rules.predicate
def is_project_contributor(user, obj):
    """Whether or not the user has the role of project contributor"""
    if not obj or not user or not user.is_authenticated:
        return False
    role_as = obj.get_role(user=user)
    return role_as and role_as.role.name == PROJECT_ROLE_CONTRIBUTOR


@rules.predicate
def is_project_guest(user, obj):
    """
    Whether or not the user has the role of project guest. Also returns true if
    project has public guest access.
    """
    if obj.get_public_access_name() == PROJECT_ROLE_GUEST:
        return True
    if not obj or not user or not user.is_authenticated:
        return False
    role_as = obj.get_role(user)
    return role_as and role_as.role.name == PROJECT_ROLE_GUEST


@rules.predicate
def is_project_viewer(user, obj):
    """
    Whether or not the user has the role of project viewer.
    """
    if obj.get_public_access_name() == PROJECT_ROLE_VIEWER:
        return True
    if not obj or not user or not user.is_authenticated:
        return False
    role_as = obj.get_role(user=user)
    return role_as and role_as.role.name == PROJECT_ROLE_VIEWER


@rules.predicate
def is_project_finder(user, obj):
    """
    Whether or not the user has the role of project finder.
    """
    if not obj or not user or not user.is_authenticated:
        return False
    role_as = obj.get_role(user)
    return role_as and role_as.role.name == PROJECT_ROLE_GUEST


@rules.predicate
def can_view_project(user, obj):
    """
    Whether or not user can view the project.
    """
    if obj.public_access:
        return True
    if user.is_authenticated and obj.has_role(user):
        return True
    return False


@rules.predicate
def has_category_child_role(user, obj):
    """
    Whether or not the user has any role in any child project under the
    current one, if the current project is a category. Also returns True if
    user is anonymous and category includes children with public guest access.
    """
    return obj.is_category() and (
        (
            user.is_authenticated
            and (obj.has_role(user) or obj.has_role_in_children(user))
        )
        or (
            user.is_anonymous
            and getattr(settings, 'PROJECTROLES_ALLOW_ANONYMOUS', False)
            and obj.has_public_children
        )
    )


@rules.predicate
def has_roles(user):
    """Whether or not the user has any roles set in the system"""
    return RoleAssignment.objects.filter(user=user).count() > 0


@rules.predicate
def is_modifiable_project(user, obj):
    """Whether or not project metadata is modifiable"""
    if obj.is_remote() or app_settings.get(APP_NAME, 'site_read_only'):
        return False
    return True


@rules.predicate
def is_public_stats_category(user, obj):
    """Whether or not project is top level category with public stats enabled"""
    if (
        obj.is_category()
        and not obj.parent
        and app_settings.get(APP_NAME, 'category_public_stats', project=obj)
    ):
        return True
    return False


@rules.predicate
def can_modify_project_data(user, obj):
    """
    Whether or not project app data can be modified, due to e.g. project
    archiving status.
    """
    return not obj.archive and not app_settings.get(APP_NAME, 'site_read_only')


@rules.predicate
def can_create_projects(user, obj):
    """Whether or not new projects can be created on the site"""
    if settings.PROJECTROLES_SITE_MODE == SITE_MODE_TARGET and (
        not settings.PROJECTROLES_TARGET_CREATE or (obj and obj.is_remote())
    ):
        return False
    if app_settings.get(APP_NAME, 'site_read_only'):
        return False
    return True


@rules.predicate
def is_allowed_anonymous(user):
    """Return True if user is anonymous and allowed by site"""
    if (
        not user
        or user.is_anonymous
        and getattr(settings, 'PROJECTROLES_ALLOW_ANONYMOUS', False)
    ):
        return True
    return False


@rules.predicate
def is_source_site():
    """Return True if PROJECTROLES_SITE_MODE is set to SITE_MODE_PROJECT"""
    return settings.PROJECTROLES_SITE_MODE == SITE_MODE_SOURCE


@rules.predicate
def is_target_site():
    """Return True if PROJECTROLES_SITE_MODE is set to SITE_MODE_TARGET"""
    return settings.PROJECTROLES_SITE_MODE == SITE_MODE_TARGET


@rules.predicate
def is_site_writable():
    """Return True if site has not been set in read-only mode"""
    return not app_settings.get(APP_NAME, 'site_read_only')


@rules.predicate
def is_project_accessible(user, obj):
    """
    Return True if project access has not been temporarily blocked by
    administrator.

    NOTE: Only needed if the view is not using ProjectPermissionMixin or
    SODARAPIProjectPermission.
    """
    return not app_settings.get(APP_NAME, 'project_access_block', project=obj)


# Combined predicates ----------------------------------------------------------


# Allow creating projects under the current category
is_project_create_user = (
    is_project_owner | is_project_delegate | is_project_contributor
)

# Allow updating project
is_project_update_user = is_project_owner | is_project_delegate

# Allow creating/updating roles
is_role_update_user = is_project_owner | is_project_delegate


# Rules ------------------------------------------------------------------------


# Rules should not be needed, use permissions for user rights


# Permissions ------------------------------------------------------------------


# Allow viewing project/category details
rules.add_perm(
    'projectroles.view_project',
    can_view_project | is_public_stats_category | has_category_child_role,
)

# Allow project updating
rules.add_perm(
    'projectroles.update_project',
    is_project_update_user & is_site_writable,
)

# Allow project deletion
rules.add_perm(
    'projectroles.delete_project',
    is_project_update_user & is_site_writable,
)

# Allow creation of projects
rules.add_perm(
    'projectroles.create_project', is_project_create_user & can_create_projects
)

# Allow viewing PROJECT scope settings
rules.add_perm('projectroles.view_project_settings', is_project_update_user)

# Allow updating project app settings
rules.add_perm(
    'projectroles.update_project_settings',
    is_project_update_user & is_modifiable_project,
)

# Allow viewing project roles
rules.add_perm(
    'projectroles.view_project_roles',
    can_view_project,
)

# Allow updating project owner
rules.add_perm(
    'projectroles.update_project_owner',
    is_project_owner & is_modifiable_project,
)

# Allow updating project delegate
rules.add_perm(
    'projectroles.update_project_delegate',
    is_project_owner & is_modifiable_project,
)

# Allow updating project members
rules.add_perm(
    'projectroles.update_project_members',
    is_role_update_user & is_modifiable_project,
)

# Allow inviting users to project via email
rules.add_perm(
    'projectroles.invite_users', is_role_update_user & is_modifiable_project
)

# Allow importing roles from another project
rules.add_perm(
    'projectroles.import_roles', is_project_owner & is_modifiable_project
)

# Allow starring/unstarring a project
rules.add_perm(
    'projectroles.star_project',
    (can_view_project | is_public_stats_category | has_category_child_role)
    & is_site_writable,
)

# Allow updating remote sites and remote project access
rules.add_perm('projectroles.update_remote', rules.is_superuser)

# Allow viewing hidden target sites
rules.add_perm(
    'projectroles.view_hidden_projects', rules.is_superuser | is_project_owner
)

# Allow updating site app settings
rules.add_perm('projectroles.update_site_settings', rules.is_superuser)



from django import template

register = template.Library()


@register.simple_tag
def get_info_cls(value: str, base_class: str = 'col-md-7') -> str:
    """Return info element class"""
    c = base_class
    if value == '':
        c += ' text-muted'
    return c


@register.simple_tag
def get_info_val(value: str) -> str:
    """Return info element value"""
    if value == '':
        value = '(Empty value)'
    return value



"""Tests for Ajax views in the adminalerts app"""

import json

from django.urls import reverse

from adminalerts.tests.test_views import AdminalertsViewTestBase


class TestAdminAlertActiveToggleAjaxView(AdminalertsViewTestBase):
    """Tests for AdminAlertActiveToggleAjaxView"""

    def setUp(self):
        super().setUp()
        self.alert = self._make_alert()
        self.url = reverse(
            'adminalerts:ajax_active_toggle',
            kwargs={'adminalert': self.alert.sodar_uuid},
        )

    def test_post_deactivate(self):
        """Test AdminAlertActiveToggleAjaxView POST to deactivate alert"""
        self.assertTrue(self.alert.active)
        with self.login(self.superuser):
            response = self.client.post(self.url)
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content)
        self.alert.refresh_from_db()
        self.assertFalse(self.alert.active)
        self.assertFalse(data['is_active'])

    def test_post_activate(self):
        """Test POST to activate alert"""
        self.alert.active = False
        self.alert.save()
        with self.login(self.superuser):
            response = self.client.post(self.url)
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content)
        self.alert.refresh_from_db()
        self.assertTrue(self.alert.active)
        self.assertTrue(data['is_active'])



"""Plugin point definitions and plugin API for apps based on projectroles"""

import json
import logging

from typing import Any, Optional, Union
from uuid import UUID

from django.conf import settings
from django.db.models import Model, QuerySet
from django.http import HttpRequest
from djangoplugins.point import PluginPoint

from projectroles.models import (
    Project,
    Role,
    RoleAssignment,
    SODARUser,
    APP_SETTING_TYPES,
    SODAR_CONSTANTS,
)


logger = logging.getLogger(__name__)


# SODAR constants
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
APP_SETTING_SCOPE_PROJECT = SODAR_CONSTANTS['APP_SETTING_SCOPE_PROJECT']
APP_SETTING_SCOPE_USER = SODAR_CONSTANTS['APP_SETTING_SCOPE_USER']
APP_SETTING_SCOPE_PROJECT_USER = SODAR_CONSTANTS[
    'APP_SETTING_SCOPE_PROJECT_USER'
]
APP_SETTING_SCOPE_SITE = SODAR_CONSTANTS['APP_SETTING_SCOPE_SITE']
APP_SETTING_TYPE_BOOLEAN = SODAR_CONSTANTS['APP_SETTING_TYPE_BOOLEAN']
APP_SETTING_TYPE_INTEGER = SODAR_CONSTANTS['APP_SETTING_TYPE_INTEGER']
APP_SETTING_TYPE_JSON = SODAR_CONSTANTS['APP_SETTING_TYPE_JSON']
APP_SETTING_TYPE_STRING = SODAR_CONSTANTS['APP_SETTING_TYPE_STRING']

# Local constants
PLUGIN_TYPES = {
    'project_app': 'ProjectAppPluginPoint',
    'backend': 'BackendPluginPoint',
    'site_app': 'SiteAppPluginPoint',
}
APP_SETTING_SCOPES = [
    APP_SETTING_SCOPE_PROJECT,
    APP_SETTING_SCOPE_USER,
    APP_SETTING_SCOPE_PROJECT_USER,
    APP_SETTING_SCOPE_SITE,
]
APP_SETTING_OPTION_TYPES = [APP_SETTING_TYPE_INTEGER, APP_SETTING_TYPE_STRING]
# TODO: Remove in v1.3 (see #1718)
PLUGIN_API_DEPRECATE_MSG = (
    'Importing and calling {method}() directly from projectroles.plugins has '
    'been deprecated and will be removed in v1.3. Use '
    'projectroles.plugins.PluginAPI instead.'
)

# From djangoplugins
ENABLED = 0
DISABLED = 1
REMOVED = 2


# Plugin Mixins ----------------------------------------------------------------


class ProjectModifyPluginMixin:
    """
    Mixin for project plugin API extensions for additional actions to be
    performed for project and role modifications. Used if e.g. updating external
    resources based on SODAR Core projects.

    Add this into your project app or backend plugin if you want to implement
    additional modification features. It is not supported on site app plugins.
    """

    def perform_project_modify(
        self,
        project: Project,
        action: str,
        project_settings: dict,
        old_data: Optional[dict] = None,
        old_settings: Optional[dict] = None,
        request: Optional[HttpRequest] = None,
    ):
        """
        Perform additional actions to finalize project creation or update.

        :param project: Current project object (Project)
        :param action: Action to perform (CREATE or UPDATE)
        :param project_settings: Project app settings (dict)
        :param old_data: Old project data in case of an update (dict or None)
        :param old_settings: Old app settings in case of update (dict or None)
        :param request: Request object or None
        """
        pass

    def revert_project_modify(
        self,
        project: Project,
        action: str,
        project_settings: dict,
        old_data: Optional[dict] = None,
        old_settings: Optional[dict] = None,
        request: Optional[HttpRequest] = None,
    ):
        """
        Revert project creation or update if errors have occurred in other apps.

        :param project: Current project object (Project)
        :param action: Action which was performed (CREATE or UPDATE)
        :param project_settings: Project app settings (dict)
        :param old_data: Old project data in case of update (dict or None)
        :param old_settings: Old app settings in case of update (dict or None)
        :param request: Request object or None
        """
        pass

    def perform_role_modify(
        self,
        role_as: RoleAssignment,
        action: str,
        old_role: Optional[Role] = None,
        request: Optional[HttpRequest] = None,
    ):
        """
        Perform additional actions to finalize role assignment creation or
        update.

        :param role_as: RoleAssignment object
        :param action: Action to perform (CREATE or UPDATE)
        :param old_role: Role object for previous role in case of an update
        :param request: Request object or None
        """
        pass

    def revert_role_modify(
        self,
        role_as: RoleAssignment,
        action: str,
        old_role: Optional[Role] = None,
        request: Optional[HttpRequest] = None,
    ):
        """
        Revert role assignment creation or update if errors have occurred in
        other apps.

        :param role_as: RoleAssignment object
        :param action: Action which was performed (CREATE or UPDATE)
        :param old_role: Role object for previous role in case of an update
        :param request: Request object or None
        """
        pass

    def perform_role_delete(
        self,
        role_as: RoleAssignment,
        request: Optional[HttpRequest] = None,
    ):
        """
        Perform additional actions to finalize role assignment deletion.

        :param role_as: RoleAssignment object
        :param request: Request object or None
        """
        pass

    def revert_role_delete(
        self,
        role_as: RoleAssignment,
        request: Optional[HttpRequest] = None,
    ):
        """
        Revert role assignment deletion deletion if errors have occurred in
        other apps.

        :param role_as: RoleAssignment object
        :param request: Request object or None
        """
        pass

    def perform_owner_transfer(
        self,
        project: Project,
        new_owner: SODARUser,
        old_owner: SODARUser,
        old_owner_role: Optional[Role] = None,
        request: Optional[HttpRequest] = None,
    ):
        """
        Perform additional actions to finalize project ownership transfer.

        :param project: Project object
        :param new_owner: SODARUser object for new owner
        :param old_owner: SODARUser object for previous owner
        :param old_owner_role: Role object for new role of old owner or None
        :param request: Request object or None
        """
        pass

    def revert_owner_transfer(
        self,
        project: Project,
        new_owner: SODARUser,
        old_owner: SODARUser,
        old_owner_role: Optional[Role] = None,
        request: Optional[HttpRequest] = None,
    ):
        """
        Revert project ownership transfer if errors have occurred in other apps.

        :param project: Project object
        :param new_owner: SODARUser object for new owner
        :param old_owner: SODARUser object for previous owner
        :param old_owner_role: Role object for new role of old owner or None
        :param request: Request object or None
        """
        pass

    def perform_project_sync(self, project: Project):
        """
        Synchronize existing projects to ensure related data exists when the
        syncmodifyapi management comment is called. Should mostly be used in
        development when the development databases have been e.g. modified or
        recreated.

        :param project: Current project object (Project)
        """
        pass

    def perform_project_setting_update(
        self,
        plugin_name: str,
        setting_name: str,
        value: Any,
        old_value: Any,
        project: Optional[Project] = None,
        user: Optional[SODARUser] = None,
    ):
        """
        Perform additional actions when updating a single app setting with
        PROJECT scope.

        :param plugin_name: Name of app plugin for the setting, "projectroles"
                            is used for projectroles settings (string)
        :param setting_name: Setting name (string)
        :param value: New value for setting
        :param old_value: Previous value for setting
        :param project: Project object or None
        :param user: User object or None
        """
        pass

    def revert_project_setting_update(
        self,
        plugin_name: str,
        setting_name: str,
        value: Any,
        old_value: Any,
        project: Optional[Project] = None,
        user: Optional[SODARUser] = None,
    ):
        """
        Revert updating a single app setting with PROJECT scope if errors have
        occurred in other apps.

        :param plugin_name: Name of app plugin for the setting, "projectroles"
                            is used for projectroles settings (string)
        :param setting_name: Setting name (string)
        :param value: New value for setting
        :param old_value: Previous value for setting
        :param project: Project object or None
        :param user: User object or None
        """
        pass

    def perform_project_archive(self, project: Project):
        """
        Perform additional actions to finalize project archiving or unarchiving.
        The state being applied can be derived from the project.archive attr.

        :param project: Project object (Project)
        """
        pass

    def revert_project_archive(self, project: Project):
        """
        Revert project archiving or unarchiving if errors have occurred in other
        apps. The state being originally set can be derived from the
        project.archive attr.

        :param project: Project object (Project)
        """
        pass

    def perform_project_delete(self, project: Project):
        """
        Perform additional actions to finalize project deletion.

        NOTE: This operation can not be undone so there is no revert method.

        :param project: Project object (Project)
        """
        pass


# Plugin Points ----------------------------------------------------------------


class ProjectAppPluginPoint(PluginPoint):
    """Projectroles plugin point for registering project specific apps"""

    #: UI URLs
    urls = []

    #: App setting definitions
    #:
    #: Example ::
    #:
    #:     app_settings = [
    #:         PluginAppSettingDef(
    #:             name='example_setting',  # Must be unique within plugin
    #:             scope=APP_SETTING_SCOPE_PROJECT,
    #:             type=APP_SETTING_TYPE_STRING,
    #:             default='example',  # Optional
    #:             label='Example setting',  # Optional
    #:             placeholder='Enter example setting here',  # Optional
    #:             description='Example user setting',  # Optional
    #:             options=['example', 'example2'],  # Optional, only for STRING
    #:                                               # or INTEGER settings
    #:             user_modifiable=True,  # Optional, show/hide in forms
    #:             global_edit=False,  # Optional, enable/disable editing on
    #:                                 # target sites
    #:             widget_attrs={},  # Optional, widget attrs for forms
    #:         )
    #:    ]
    app_settings = []

    # DEPRECATED, will be removed in the next SODAR Core release
    project_settings = {}

    #: Iconify icon
    icon = 'mdi:help-rhombus-outline'

    #: Entry point URL ID (must take project sodar_uuid as "project" argument)
    entry_point_url_id = 'home'

    #: Description string
    description = 'TODO: Write a description for your plugin'

    #: Required permission for accessing the app
    app_permission = None

    #: Enable or disable general search from project title bar
    search_enable = False

    #: List of search object types for the app
    search_types = []

    #: Search results template
    search_template = None

    #: App card template for the project details page
    details_template = None

    #: App card title for the project details page
    details_title = None

    #: Position in plugin ordering
    plugin_ordering = 50

    #: Optional custom project list column definition
    #:
    #: Example ::
    #:
    #:     project_list_columns = {
    #:         'column_id': {
    #:             'title': 'Column Title',
    #:             'width': 100,  # Desired width of column in pixels
    #:             'description': 'Description',  # Optional description string
    #:             'active': True,  # Boolean for whether the column is active
    #:             'ordering': 50,  # Integer for ordering the columns
    #:             'align': 'left'  # Alignment of content
    #:         }
    #:     }
    project_list_columns = {}

    #: Display application for categories in addition to projects
    category_enable = False

    #: Names of plugin specific Django settings to display in siteinfo
    info_settings = []

    def get_object(
        self, model: type[Model], uuid: Union[str, UUID]
    ) -> Optional[Model]:
        """
        Return object based on a model class and the object's SODAR UUID.

        :param model: Object model class
        :param uuid: sodar_uuid of the referred object
        :return: Model object or None if not found
        :raise: NameError if model is not found
        """
        try:
            return model.objects.get(sodar_uuid=uuid)
        except model.DoesNotExist:
            return None

    def get_object_link(
        self, model_str: str, uuid: Union[str, UUID]
    ) -> Optional['PluginObjectLink']:
        """
        Return URL referring to an object used by the app, along with a name to
        be shown to the user for linking.

        :param model_str: Object class (string)
        :param uuid: sodar_uuid of the referred object
        :return: PluginObjectLink or None if not found
        """
        obj = self.get_object(eval(model_str), uuid)
        if not obj:
            return None
        return None

    def get_extra_data_link(
        self, _extra_data: dict, _name: str
    ) -> Optional[str]:
        """Return a link for timeline label starting with 'extra-'"""
        return None

    def search(
        self,
        search_terms: list[str],
        user: SODARUser,
        search_type: Optional[str] = None,
        keywords: Optional[list[str]] = None,
    ) -> list['PluginSearchResult']:
        """
        Return app items based on one or more search terms, user, optional type
        and optional keywords.

        :param search_terms: Search terms to be joined with the OR operator
                             (list of strings)
        :param user: User object for user initiating the search
        :param search_type: String
        :param keywords: List (optional)
        :return: List of PluginSearchResult objects
        """
        # TODO: If implemented, also implement display of results in the app's
        #       search template
        return []

    def update_cache(
        self,
        name: Optional[str] = None,
        project: Optional[Project] = None,
        user: Optional[SODARUser] = None,
    ):
        """
        Update cached data for this app, limitable to item ID and/or project.

        :param name: Item name to limit update to (string, optional)
        :param project: Project object to limit update to (optional)
        :param user: User object to denote user triggering the update (optional)
        """
        return None

    def get_statistics(self) -> dict:
        """
        Return app statistics as a dict. Should take the form of
        {id: {label, value, url (optional), description (optional)}}.

        :return: Dict
        """
        return {}

    def get_category_stats(
        self, category: Project
    ) -> list['PluginCategoryStatistic']:
        """
        Return app statistics for the given category. Expected to return
        cumulative statistics for all projects under the category and its
        possible subcategories.

        :param category: Project object of CATEGORY type
        :return: List of PluginCategoryStatistic objects
        """
        return []

    def get_project_list_value(
        self, column_id: str, project: Project, user: SODARUser
    ) -> Union[str, int, None]:
        """
        Return a value for the optional additional project list column specific
        to a project.

        :param column_id: ID of the column (string)
        :param project: Project object
        :param user: User object (current user)
        :return: String (may contain HTML), integer or None
        """
        return None

    def validate_form_app_settings(
        self,
        app_settings: dict,
        project: Optional[Project] = None,
        user: Optional[SODARUser] = None,
    ) -> Optional[dict]:
        """
        Validate app settings form data and return a dict of errors.

        :param app_settings: Dict of app settings
        :param project: Project object or None
        :param user: User object or None
        :return: dict in format of {setting_name: 'Error string'}
        """
        return None


class BackendPluginPoint(PluginPoint):
    """Projectroles plugin point for registering backend apps"""

    #: Iconify icon
    icon = 'mdi:help-rhombus-outline'

    #: Description string
    description = 'TODO: Write a description for your plugin'

    #: URL of optional javascript file to be included
    javascript_url = None

    #: URL of optional css file to be included
    css_url = None

    #: Names of plugin specific Django settings to display in siteinfo
    info_settings = []

    def get_api(self) -> Any:
        """Return API entry point object"""
        raise NotImplementedError

    def get_statistics(self) -> dict:
        """
        Return backend statistics as a dict. Should take the form of
        {id: {label, value, url (optional), description (optional)}}.

        :return: Dict
        """
        return {}

    def get_object(
        self, model: type[Model], uuid: Union[str, UUID]
    ) -> Optional[Model]:
        """
        Return object based on a model class and the object's SODAR UUID.

        :param model: Object model class
        :param uuid: sodar_uuid of the referred object
        :return: Model object or None if not found
        :raise: NameError if model is not found
        """
        try:
            return model.objects.get(sodar_uuid=uuid)
        except model.DoesNotExist:
            return None

    def get_object_link(
        self, model_str: str, uuid: Union[str, UUID]
    ) -> Optional['PluginObjectLink']:
        """
        Return URL referring to an object used by the app, along with a name to
        be shown to the user for linking.

        :param model_str: Object class (string)
        :param uuid: sodar_uuid of the referred object
        :return: PluginObjectLink or None if not found
        """
        obj = self.get_object(eval(model_str), uuid)
        if not obj:
            return None
        return None

    def get_extra_data_link(
        self, _extra_data: dict, _name: str
    ) -> Optional[str]:
        """Return a link for timeline label starting with 'extra-'"""
        return None


class SiteAppPluginPoint(PluginPoint):
    """Projectroles plugin point for registering site-wide apps"""

    #: Iconify icon
    icon = 'mdi:help-rhombus-outline'

    #: Description string
    description = 'TODO: Write a description for your plugin'

    #: Entry point URL ID
    entry_point_url_id = 'home'

    #: Required permission for displaying the app
    app_permission = None

    #: User settings definition
    #:
    #: Example ::
    #:
    #:     app_settings = [
    #:         PluginAppSettingDef(
    #:             name='example_setting',  # Must be unique within plugin
    #:             scope=APP_SETTING_SCOPE_USER,  # Use USER or SITE
    #:             type=APP_SETTING_TYPE_STRING,
    #:             default='example',  # Optional
    #:             label='Example setting',  # Optional
    #:             placeholder='Enter example setting here',  # Optional
    #:             description='Example user setting',  # Optional
    #:             options=['example', 'example2'],  # Optional, only for STRING
    #:                                               # or INTEGER settings
    #:             user_modifiable=True,  # Optional, show/hide in forms
    #:             global_edit=False,  # Optional, enable/disable editing on
    #:                                 # target sites
    #:             widget_attrs={},  # Optional, widget attrs for forms
    #:         )
    #:    ]
    app_settings = []

    #: List of names for plugin specific Django settings to display in siteinfo
    info_settings = []

    def get_statistics(self) -> dict:
        """
        Return app statistics as a dict. Should take the form of
        {id: {label, value, url (optional), description (optional)}}.

        :return: Dict
        """
        return {}

    def get_messages(self, user: Optional[SODARUser] = None) -> list[dict]:
        """
        Return a list of messages to be shown to users.

        :param user: User object (optional)
        :return: List of dicts or and empty list if no messages
        """
        '''
        Example of valid return data:
        return [{
            'content': 'Message content in here, can contain html',
            'color': 'info',        # Corresponds to bg-* in Bootstrap
            'dismissable': True     # False for non-dismissable
        }]
        '''
        return []

    def get_object(
        self, model: type[Model], uuid: Union[str, UUID]
    ) -> Optional[Model]:
        """
        Return object based on a model class and the object's SODAR UUID.

        :param model: Object model class
        :param uuid: sodar_uuid of the referred object
        :return: Model object or None if not found
        :raise: NameError if model is not found
        """
        try:
            return model.objects.get(sodar_uuid=uuid)
        except model.DoesNotExist:
            return None

    def get_object_link(
        self, model_str: str, uuid: Union[str, UUID]
    ) -> Optional['PluginObjectLink']:
        """
        Return URL referring to an object used by the app, along with a name to
        be shown to the user for linking.

        :param model_str: Object class (string)
        :param uuid: sodar_uuid of the referred object
        :return: PluginObjectLink or None if not found
        """
        obj = self.get_object(eval(model_str), uuid)
        if not obj:
            return None
        return None

    def get_extra_data_link(
        self, _extra_data: dict, _name: str
    ) -> Optional[str]:
        """Return a link for timeline label starting with 'extra-'"""
        return None

    def validate_form_app_settings(
        self,
        app_settings: dict,
        user: Optional[SODARUser] = None,
    ) -> Optional[dict]:
        """
        Validate app settings form data and return a dict of errors.

        :param app_settings: Dict of app settings
        :param user: User object or None
        :return: dict in format of {setting_name: 'Error string'}
        """
        return None


# Data Classes -----------------------------------------------------------------


class PluginAppSettingDef:
    """
    Class representing an AppSetting definition. Expected to be used to define
    app settings in the plugin app_settings variable.
    """

    def __init__(
        self,
        name: str,
        scope: str,
        type: str,
        default: Optional[Any] = None,
        label: Optional[str] = None,
        placeholder: Union[str, int, None] = None,
        description: Optional[str] = None,
        options: Union[callable, list, None] = None,
        user_modifiable: Optional[
            bool
        ] = None,  # Set None here to validate PROJECT_USER value
        global_edit: bool = False,
        project_types: Optional[list[str]] = None,
        widget_attrs: Optional[dict] = None,
    ):
        """
        Initialize PluginAppSettingDef.

        :param name: Setting name to be used internally (string)
        :param scope: Setting scope, must correspond to one of
                      APP_SETTING_SCOPE_* (string)
        :param type: Setting type, must correspond to one of APP_SETTING_TYPE_*
                     (string)
        :param default: Default value, type depends on setting type. Can be a
                        callable.
        :param label: Display label (string, optional, name is used as default)
        :param placeholder: Placeholder value to be displayed in forms (string,
                            optional)
        :param description: Detailed setting description (string, optional)
        :param options: Limit value to given options. Can be callable (optional,
                        only for STRING or INTEGER types)
        :param user_modifiable: Display in forms for user if True (optional,
                                default=True, only for PROJECT and USER scopes)
        :param global_edit: Only allow editing on source site if True (optional,
                            default=False)
        :param project_types: Allowed project types (optional,
                              default=[PROJECT_TYPE_PROJECT])
        :parm widget_attrs: Form widget attributes (optional, dict)
        :raise: ValueError if an argument is not valid
        """
        # Validate provided values
        self.validate_scope(scope)
        self.validate_type(type)
        self.validate_type_options(type, options)
        if not callable(default):
            self.validate_value(type, default)
        if (
            default is not None
            and options is not None
            and not callable(default)
            and not callable(options)
        ):
            self.validate_default_in_options(default, options)
        # Validate and set user_modifiable
        self.validate_user_modifiable(scope, user_modifiable)
        if user_modifiable is None:
            user_modifiable = (
                False if scope == APP_SETTING_SCOPE_PROJECT_USER else True
            )
        # Set members
        self.name = name
        self.scope = scope
        self.type = type
        self.default = (
            ''
            if type == APP_SETTING_TYPE_STRING and default is None
            else default
        )
        self.label = label
        self.placeholder = placeholder
        self.description = description
        self.options = options or []
        self.user_modifiable = user_modifiable
        self.global_edit = global_edit
        self.project_types = project_types or [PROJECT_TYPE_PROJECT]
        self.widget_attrs = widget_attrs or {}

    @classmethod
    def validate_scope(cls, scope: str):
        """
        Validate the app setting scope.

        :param scope: String
        :raise: ValueError if scope is not recognized
        """
        if scope not in APP_SETTING_SCOPES:
            raise ValueError(f'Invalid scope "{scope}"')

    @classmethod
    def validate_type(cls, setting_type: str):
        """
        Validate the app setting type.

        :param setting_type: String
        :raise: ValueError if type is not recognized
        """
        if setting_type not in APP_SETTING_TYPES:
            raise ValueError(f'Invalid setting type "{setting_type}"')

    @classmethod
    def validate_type_options(cls, setting_type: str, setting_options: list):
        """
        Validate existence of options against setting type.

        :param setting_type: String
        :param setting_options: List of options (Strings or Integers)
        :raise: ValueError if type is not recognized
        """
        if (
            setting_type
            not in [APP_SETTING_TYPE_INTEGER, APP_SETTING_TYPE_STRING]
            and setting_options
        ):
            raise ValueError(
                'Options are only allowed for settings of type INTEGER and '
                'STRING'
            )

    @classmethod
    def validate_default_in_options(
        cls, setting_default: Any, setting_options: Any
    ):
        """
        Validate existence of default value in uncallable options.

        :param setting_default: Default value
        :param setting_options: Setting options
        :raise: ValueError if default is not found in options
        """
        opts = [o[0] if isinstance(o, tuple) else o for o in setting_options]
        if (
            setting_options is not None
            and not callable(setting_options)
            and setting_default is not None
            and not callable(setting_default)
            and setting_default not in opts
        ):
            raise ValueError(
                'Default value "{}" not found in options ({})'.format(
                    setting_default,
                    ', '.join([str(o) for o in opts]),
                )
            )

    @classmethod
    def validate_value(cls, setting_type: str, setting_value: Any):
        """
        Validate non-callable value.

        :param setting_type: Setting type (string)
        :param setting_value: Setting value
        :raise: ValueError if value is invalid
        """
        if setting_type == APP_SETTING_TYPE_BOOLEAN:
            if not isinstance(setting_value, bool):
                raise ValueError(
                    f'Please enter value as bool (value={setting_value}; '
                    f'type={type(setting_value)})'
                )
        elif setting_type == APP_SETTING_TYPE_INTEGER:
            if (
                not isinstance(setting_value, int)
                and not str(setting_value).isdigit()
            ):
                raise ValueError(
                    f'Please enter a valid integer value ({setting_value})'
                )
        elif setting_type == APP_SETTING_TYPE_JSON:
            if setting_value and not isinstance(setting_value, (dict, list)):
                raise ValueError(
                    f'Please enter JSON value as dict or list '
                    f'(value={setting_value}; type={type(setting_value)})'
                )
            try:
                json.dumps(setting_value)
            except TypeError:
                raise ValueError(f'Please enter valid JSON ({setting_value})')

    @classmethod
    def validate_user_modifiable(
        cls, scope: str, user_modifiable: Optional[bool]
    ):
        """
        Validate user_modifiable against scope.

        :param scope: String
        :param user_modifiable: Bool or None
        :raises: ValueError if user_modifiable can't be set for scope
        """
        if user_modifiable and scope == APP_SETTING_SCOPE_PROJECT_USER:
            raise ValueError(
                'Argument user_modifiable can not be set True for PROJECT_USER '
                'scope settings'
            )


class PluginObjectLink:
    """
    Class representing a hyperlink to an object used by the app. Expected to be
    returned from get_object_link() implementations.
    """

    #: URL to the object (string)
    url = None
    #: Name of the object to be displayed in link (string, formerly "label")
    name = None
    #: Open the link in a blank browser tab (boolean, default=False)
    blank = False

    def __init__(self, url: str, name: str, blank: bool = False):
        """
        Initialize PluginObjectLink.

        :param url: URL to the object (string)
        :param name: Name of the object to be displayed in link (string,
                     formerly "label")
        :param blank: Open the link in a blank browser tab (boolean,
                      default=False)
        """
        self.url = url
        self.name = name
        self.blank = blank


class PluginSearchResult:
    """
    Class representing a list of search results from a specific plugin for one
    or more search types. Expected to be returned from search() implementations.
    """

    #: Category of the result set, used in templates (string)
    category = None
    #: Title to be displayed for this set of search results in the UI (string)
    title = None
    #: List of one or more search type keywords for these results
    search_types = []
    #: List or QuerySet of result objects
    items = []

    def __init__(
        self,
        category: str,
        title: str,
        search_types: Optional[list[str]],
        items: Union[list, QuerySet],
    ):
        """
        Initialize PluginSearchResult.

        :param category: Category of the result set, used in templates (string)
        :param title: Title to be displayed for this set of search results in
                      the UI (string)
        :param search_types: List of one or more search type keywords for the
                             results
        :param items: List or QuerySet of result objects
        """
        self.category = category
        self.title = title
        self.search_types = search_types
        if not isinstance(search_types, list) or len(search_types) < 1:
            raise ValueError(
                'At least one type keyword must be provided in search_types'
            )
        self.items = items


class PluginCategoryStatistic:
    """Class representing an item of category statistics"""

    #: Plugin from which this item is initialized (None for projectroles)
    plugin = None
    #: Title to be displayed (string)
    title = None
    #: Value of the statistic (int or float)
    value = 0
    #: Optional unit (string or None)
    unit = None
    #: Optional description to be displayed as tooltip (string or None)
    description = None
    #: Optional icon namespace and ID (string or None)

    def __init__(
        self,
        plugin: Optional[ProjectAppPluginPoint],
        title: str,
        value: Union[int, float],
        unit: Optional[str] = None,
        description: Optional[str] = None,
        icon: Optional[str] = None,
    ):
        """
        Initialize PluginCategoryStatistic.

        :param plugin: ProjectAppPluginPoint object (None for projectroles)
        :param title: String
        :param value: Integer or float
        :param unit: String or None
        :param description: String or None
        :param icon: String or None
        :raises: ValueError if value is not integer or float
        """
        self.plugin = plugin
        self.title = title
        if not isinstance(value, int) and not isinstance(value, float):
            raise ValueError('Value must be integer or float')
        self.value = value
        self.unit = unit
        self.description = description
        if icon:
            self.icon = icon
        elif plugin:
            self.icon = getattr(plugin, 'icon')


# Plugin API -------------------------------------------------------------------


class PluginAPI:
    """API for SODAR Core plugin retrieval"""

    @classmethod
    def get_active_plugins(
        cls, plugin_type: str = 'project_app', custom_order: bool = False
    ) -> Optional[list]:
        """
        Return active plugins of a specific type.

        :param plugin_type: "project_app", "site_app" or "backend" (string)
        :param custom_order: Order by plugin_ordering for project apps (boolean)
        :return: List or None
        :raise: ValueError if plugin_type is not recognized
        """
        if plugin_type not in PLUGIN_TYPES.keys():
            raise ValueError(
                'Invalid value for plugin_type. Accepted values: {}'.format(
                    ', '.join(PLUGIN_TYPES.keys())
                )
            )
        plugins = eval(PLUGIN_TYPES[plugin_type]).get_plugins()
        if plugins:
            ret = [
                p
                for p in plugins
                if (
                    p.is_active()
                    and (
                        plugin_type in ['project_app', 'site_app']
                        or p.name in settings.ENABLED_BACKEND_PLUGINS
                    )
                )
            ]
            return sorted(
                ret,
                key=lambda x: (
                    x.plugin_ordering
                    if custom_order and plugin_type == 'project_app'
                    else x.name
                ),
            )
        return None

    @classmethod
    def change_plugin_status(
        cls, name: str, status: int, plugin_type: str = 'app'
    ):
        """
        Change the status of a selected plugin in the database.

        :param name: Plugin name (string)
        :param status: Status (int, see djangoplugins)
        :param plugin_type: Type of plugin ("app", "backend" or "site")
        :raise: ValueError if plugin_type is invalid or plugin with name not
                found
        """
        # NOTE: Used to forge plugin to a specific status for e.g. testing
        if plugin_type == 'app':
            plugin = ProjectAppPluginPoint.get_plugin(name)
        elif plugin_type == 'backend':
            plugin = BackendPluginPoint.get_plugin(name)
        elif plugin_type == 'site':
            plugin = SiteAppPluginPoint.get_plugin(name)
        else:
            raise ValueError(f'Invalid plugin_type: "{plugin_type}"')
        if not plugin:
            raise ValueError(
                f'Plugin of type "{plugin_type}" not found with name "{name}"'
            )
        plugin = plugin.get_model()
        plugin.status = status
        plugin.save()

    @classmethod
    def get_app_plugin(
        cls, plugin_name: str, plugin_type: Optional[str] = None
    ) -> Any:
        """
        Return active app plugin.

        :param plugin_name: Plugin name (string)
        :param plugin_type: Plugin type (string or None for all types)
        :return: Plugin object or None if not found or inactive
        """
        if plugin_type:
            plugin_types = [PLUGIN_TYPES[plugin_type]]
        else:
            plugin_types = PLUGIN_TYPES.values()
        for t in plugin_types:
            try:
                return eval(t).get_plugin(plugin_name)
            except Exception:
                pass
        return None

    @classmethod
    def get_backend_api(
        cls, plugin_name: str, force: bool = False, **kwargs
    ) -> Any:
        """
        Return backend API object.
        NOTE: May raise an exception from plugin.get_api().

        :param plugin_name: Plugin name (string)
        :param force: Return plugin regardless of status in
                      ENABLED_BACKEND_PLUGINS
        :param kwargs: Optional kwargs for API
        :return: Plugin object or None if not found
        """
        if plugin_name in settings.ENABLED_BACKEND_PLUGINS or force:
            try:
                plugin = BackendPluginPoint.get_plugin(plugin_name)
            except BackendPluginPoint.DoesNotExist:
                return None
            return plugin.get_api(**kwargs) if plugin.is_active() else None
        return None


# TODO: Remove in v1.3 (see #1718)
def get_active_plugins(
    plugin_type: str = 'project_app', custom_order: bool = False
) -> Optional[list[PluginPoint]]:
    """
    Return active plugins of a specific type.

    DEPRECATED: To be removed in v1.3. Use method in PluginAPI instead.

    :param plugin_type: "project_app", "site_app" or "backend" (string)
    :param custom_order: Order by plugin_ordering for project apps (boolean)
    :return: List or None
    :raise: ValueError if plugin_type is not recognized
    """
    logger.warning(PLUGIN_API_DEPRECATE_MSG.format(method='get_active_plugins'))
    plugin_api = PluginAPI()
    return plugin_api.get_active_plugins(plugin_type, custom_order)


# TODO: Remove in v1.3 (see #1718)
def change_plugin_status(name: str, status: int, plugin_type: str = 'app'):
    """
    Change the status of a selected plugin in the database.

    DEPRECATED: To be removed in v1.3. Use method in PluginAPI instead.

    :param name: Plugin name (string)
    :param status: Status (int, see djangoplugins)
    :param plugin_type: Type of plugin ("app", "backend" or "site")
    :raise: ValueError if plugin_type is invalid or plugin with name not found
    """
    # NOTE: Used to forge plugin to a specific status for e.g. testing
    logger.warning(
        PLUGIN_API_DEPRECATE_MSG.format(method='change_plugin_status')
    )
    plugin_api = PluginAPI()
    plugin_api.change_plugin_status(name, status, plugin_type)


# TODO: Remove in v1.3 (see #1718)
def get_app_plugin(plugin_name: str, plugin_type: Optional[str] = None) -> Any:
    """
    Return active app plugin.

    DEPRECATED: To be removed in v1.3. Use method in PluginAPI instead.

    :param plugin_name: Plugin name (string)
    :param plugin_type: Plugin type (string or None for all types)
    :return: Plugin object or None if not found or inactive
    """
    logger.warning(PLUGIN_API_DEPRECATE_MSG.format(method='get_app_plugin'))
    plugin_api = PluginAPI()
    return plugin_api.get_app_plugin(plugin_name, plugin_type)


# TODO: Remove in v1.3 (see #1718)
def get_backend_api(plugin_name: str, force: bool = False, **kwargs) -> Any:
    """
    Return backend API object.
    NOTE: May raise an exception from plugin.get_api().

    DEPRECATED: To be removed in v1.3. Use method in PluginAPI instead.

    :param plugin_name: Plugin name (string)
    :param force: Return plugin regardless of status in ENABLED_BACKEND_PLUGINS
    :param kwargs: Optional kwargs for API
    :return: Plugin object or None if not found
    """
    logger.warning(PLUGIN_API_DEPRECATE_MSG.format(method='get_backend_api'))
    plugin_api = PluginAPI()
    return plugin_api.get_backend_api(plugin_name, force, **kwargs)


# Plugins within projectroles --------------------------------------------------


class RemoteSiteAppPlugin(SiteAppPluginPoint):
    """Site plugin for remote site and project management"""

    #: Name (used as plugin ID)
    name = 'remotesites'

    #: Title (used in templates)
    title = 'Remote Site Access'

    #: UI URLs
    urls = []

    #: Iconify icon
    icon = 'mdi:cloud-sync'

    #: Description string
    description = 'Management of remote SODAR sites and remote project access'

    #: Entry point URL ID
    entry_point_url_id = 'projectroles:remote_sites'

    #: Required permission for displaying the app
    app_permission = 'projectroles.update_remote'


class SiteAppSettingsAppPlugin(SiteAppPluginPoint):
    """Site plugin for site app settings"""

    #: Name (used as plugin ID)
    name = 'siteappsettings'

    #: Title (used in templates)
    title = 'Site App Settings'

    #: UI URLs
    urls = []

    #: Iconify icon
    icon = 'mdi:cog-outline'

    #: Description string
    description = 'Site-wide app setting management'

    #: Entry point URL ID
    entry_point_url_id = 'projectroles:site_app_settings'

    #: Required permission for displaying the app
    app_permission = 'projectroles.update_site_settings'
